/** @preserve Eloquence WEBDLG v2 Plugin Example: Editgrid
 (C) Copyright Marxmeier Software AG, 2023
 $Id: webdlg-editgrid.js,v 28.6 2023/11/28 09:06:17 mike Exp $
*/
eq.plugin("Editgrid", {
   // Attribute descriptor.
   atr : [
      [ 'columns',      eq.atrGET|eq.atrSET          |eq.atrINT|eq.atrPRI ],
      [ 'delim',        eq.atrGET|eq.atrSET          |eq.atrSTR|eq.atrPRI ],
      [ 'activecolumn', eq.atrGET|eq.atrSET          |eq.atrINT ],
      [ 'activeline',   eq.atrGET|eq.atrSET          |eq.atrINT ],
      [ 'cx',           eq.atrGET|eq.atrSET          |eq.atrINT ],
      [ 'cy',           eq.atrGET|eq.atrSET          |eq.atrINT ],
      [ 'colreadonly',  eq.atrGET|eq.atrSET|eq.atrIDX|eq.atrINT ],
      [ 'colwidth',     eq.atrGET|eq.atrSET|eq.atrIDX|eq.atrINT ],
      [ 'header',       eq.atrGET|eq.atrSET|eq.atrIDX|eq.atrSTR ],
      [ 'add',                    eq.atrSET|eq.atrLST|eq.atrSTR|eq.atrNOC ],
      [ 'cell',         eq.atrGET|eq.atrSET|eq.atrIDX|eq.atrSTR|eq.atrNOC ],
      [ 'clear',                  eq.atrSET          |eq.atrINT
                                                     |eq.atrSTR|eq.atrNOC ],
      [ 'content',      eq.atrGET|eq.atrSET          |eq.atrSTR|eq.atrNOC ],
      [ 'del',                    eq.atrSET|eq.atrIDX|eq.atrINT|eq.atrNOC ],
      [ 'insert',                 eq.atrSET|eq.atrIDX|eq.atrSTR|eq.atrNOC ],
      [ 'line',         eq.atrGET|eq.atrSET|eq.atrIDX|eq.atrSTR|eq.atrNOC ]
   ],

   // Class properties.
   prp : {
      minVersion  : 1,
      classList   : 'editgrid',
      changeEvent : 'mousedown input change',
      actionEvent : 'blur',
      border      : true
   },

   // API: Constructor.
   apiInit : function(dlg, el) {
      // Object state.
      this.disabled = false;
      this.cols = 0;
      this.width = [ null ];
      this.readOnly = [];
      this.hidden = [];
      this.delim = '\t';
      this.cx = 0;
      this.cy = 0;
      this.cursorPending = undefined;
      this.cursorUpdated = false;
      this.contentChanged = false;

      // Setup DOM element.
      var c, d;
      // Title bar, initially empty.
      el.appendChild(c = document.createElement('DIV'));
      c.className = 'hdr empty';
      // Title col#0: row-titles.
      c.appendChild(d = document.createElement('DIV'));
      d.className = 'r cell';
      d.appendChild(document.createElement('SPAN'));
      // Title col#1..: columns.
      c.appendChild(d = document.createElement('DIV'));
      d.className = 'c';
      // Rows.
      el.appendChild(c = document.createElement('DIV'));
      c.className = 'row';
      // Col#0: Row-titles.
      c.appendChild(d = document.createElement('DIV'));
      d.className = 'r';
      // Col#1..: Columns.
      c.appendChild(d = document.createElement('DIV'));
      d.className = 'c';
      d.onscroll = this.constructor.onScrollColumns;
   },

   // API: Set attribute value.
   apiSetAttr : function(dlg, id, el, atr, idx, val) {
      var
         cls = this.constructor,
         ix = idx ? idx.split('][') : null;
      switch (atr) {
         case 'font': {
            // See 10-dlg.js.
            if (val.fontRules)
               this.setCssRule(id, cls.cssRule.font, ">*" + val.fontRules);
            this.setCssRule(id, cls.cssRule.height,
               " .cell>*{height:" + cls.cssHeight(dlg, 1) + "}");
            var cw;
            if (this.width[0] === null) {
               if (cw === undefined)
                  cw = cls.cssWidth(dlg, 5);
               this.setCssRule(id, cls.cssRule.width, [
                  ">.hdr>.r>*,",
                  ">.row>.r>*>*{width:" + cw + "}"
               ]);
            }
            for (var i = 1, l = this.cols; i <= l; i++)
               if (this.width[i] === null) {
                  if (cw === undefined)
                     cw = cls.cssWidth(dlg, 5);
                  this.setCssRule(id, cls.cssRule.width+i, [
                     ">.hdr>.c>*:nth-child("+i+")>*,",
                     ">.row>.c>*:nth-child("+i+")>*>*{width:" + cw + "}"
                  ]);
               }
            break;
         }

         case 'columns':
            if (!ix)
               this.setColumns(dlg, id, el, val.i);
            break;

         case 'delim':
            if (!ix)
               this.delim = val.s.length ? val.s : '\t';
            break;

         case 'activecolumn':
         case 'cx':
            if (!ix)
               this.selectCell(el, val.i, this.cy);
            break;

         case 'activeline':
         case 'cy':
            if (!ix)
               this.selectCell(el, this.cx, val.i);
            break;

         case 'colreadonly':
            if (ix && ix.length === 1)
               this.setReadOnly(el, +ix[0], val.i != 0);
            break;

         case 'colwidth':
            if (ix && ix.length === 1)
               this.setWidth(dlg, id, el, +ix[0], val.i);
            break;

         case 'header':
            if (ix && ix.length === 1)
               this.setTitle(el, +ix[0], val.s);
            break;

         case 'add':
            if (!ix)
               this.addRow(el, null, val.s);
            break;

         case 'cell':
            if (ix && ix.length === 2)
               this.setCell(el, +ix[0], +ix[1], val.s);
            break;

         case 'clear':
            if (!ix)
               cls.clearRows(el);
            break;

         case 'content':
            if (!ix) {
               cls.clearRows(el);
               this.addRow(el, null, val.s);
            }
            break;

         case 'del':
            if (ix && ix.length === 1)
               cls.delRows(el, +ix[0], val.i);
            break;

         case 'insert':
            if (ix && ix.length === 1)
               this.addRow(el, +ix[0], val.s);
            break;

         case 'line':
            if (ix && ix.length === 1)
               this.setRow(el, +ix[0], val.s);
      }
   },

   // API: Get attribute value.
   apiGetAttr : function(dlg, id, el, atr, idx) {
      var ix = idx ? idx.split('][') : null;
      switch (atr) {
         case 'cell':
            if (ix && ix.length === 2)
               return this.getCell(el, +ix[0], +ix[1]);
            break;

         case 'content':
            if (!ix)
               return this.getContent(el);
            break;

         case 'line':
            if (ix && ix.length === 1)
               return this.getRow(el, +ix[0]);
      }
      return undefined;
   },

   // API: Set element-specific 'sensitive' properties.
   apiSetSensitive : function(dlg, id, el, sensitive) {
      this.disabled = !sensitive;
      var n = el.getElementsByTagName('INPUT');
      for (var i = 0; i < n; i++)
         n[i].disabled = !sensitive;
   },

   // API: Filter and/or handle key event.
   apiOnKey : function(dlg, id, el, e) {
      var key, tab, input;
      switch (key = e.key) {
         case 'Enter':
         case 'Tab':
            tab = e.shiftKey ? -1 : 1;
            break;
         case 'Home':
         case 'End':
         case 'PageUp':
         case 'PageDown':
            break;
         case 'ArrowUp':
         case 'Up':
            key = 'Up';
            break;
         case 'ArrowDown':
         case 'Down':
            key = 'Down';
            break;
         case 'ArrowLeft':
         case 'Left':
            key = 'Left';
            break;
         case 'ArrowRight':
         case 'Right':
            key = 'Right';
            break;
         default:
            return false; // Pass event to default handler.
      }
      if (!(input = this.inputCell(el))) {
         // No cell selected, select first cell.
         this.cx = this.cy = 0;
         input = this.nextCell(el);
      }
      else if (tab) {
         var
            cx = this.cx,
            cy = this.cy;
         if (tab === -1)
            input = this.prevCell(el);
         else
            input = this.nextCell(el);
         if (!input)
            return false; // Pass event to default handler.
         if (this.contentChanged) {
            this.contentChanged = false;
            if (el.classList.contains('eq-rule')) {
               this.cursorPending = {
                  cx : this.cx,
                  cy : this.cy
               };
               this.cx = cx;
               this.cy = cy;
               this.submit(el.id, el, 'blur');
               return true; // Consume event.
            }
         }
      }
      else {
         var
            cols = el.querySelector('.row>.c').children,
            cols_l = cols.length,
            rows_l = cols[0].children.length,
            cx, cy, readOnly, hidden;
         if (rows_l && cols_l) {
            cx = this.cx;
            cy = this.cy;
            readOnly = this.readOnly;
            hidden = this.hidden;
            switch (key) {
               case 'Home':
                  for (cx = 0; cx < cols_l; cx++)
                     if (!readOnly[cx] && !hidden[cx])
                        break;
                  if (cx++ == cols_l)
                     return true; // Consume event.
                  cy = 1;
                  break;
               case 'End':
                  for (cx = cols_l; --cx >= 0;)
                     if (!readOnly[cx] && !hidden[cx])
                        break;
                  if (cx++ < 0)
                     return true; // Consume event.
                  cy = rows_l;
                  break;
               case 'PageUp':
                  cy = 1;
                  break;
               case 'PageDown':
                  cy = rows_l;
                  break;
               case 'Up':
                  if (cy > 1)
                     cy--;
                  break;
               case 'Down':
                  if (cy < rows_l)
                     cy++;
                  break;
               case 'Left':
                  if (input.selectionStart !== 0)
                     return null; // Delegate event to target.
                  cx--;
                  while (--cx >= 0)
                     if (!readOnly[cx] && !hidden[cx])
                        break;
                  if (cx++ < 0)
                     return true; // Consume event.
                  break;
               case 'Right':
                  if (input.selectionStart < input.value.length)
                     return null; // Delegate event to target.
                  for (; cx < cols_l; cx++)
                     if (!readOnly[cx] && !hidden[cx])
                        break;
                  if (cx++ == cols_l)
                     return true; // Consume event.
            }
            if (this.cx !== cx || this.cy !== cy) {
               if (this.contentChanged) {
                  this.contentChanged = false;
                  if (el.classList.contains('eq-rule')) {
                     this.cursorPending = {
                        cx : cx,
                        cy : cy
                     };
                     this.submit(el.id, el, 'blur');
                     return true; // Consume event.
                  }
               }
               this.cx = cx;
               this.cy = cy;
               if (!(input = this.inputCell(el)))
                  throw new Error("Editgrid.onKey BUG: cx=" + cx + " cy=" + cy);
            }
         }
      }
      if (input) {
         this.cursorChanged(id);
         this.constructor.scrollCellIntoView(input);
         input.select();
         input.scrollLeft = input.scrollWidth;
         input.focus();
      }
      return true; // Consume event.
   },

   // API: Process type-ahead key stroke.
   apiOnTypeAhead : function(dlg, id, el, key) {
      if (key.length === 1) {
         var
            input = this.inputCell(el),
            v, s0, s1;
         if (input) {
            v = input.value;
            s0 = input.selectionStart;
            s1 = input.selectionEnd;
            if (s0 > 0) {
               if (s1 < v.length)
                  v = v.substring(0, s0) + key + v.substring(s1);
               else
                  v = v.substring(0, s0) + key;
            }
            else if (s1 < v.length)
               v = key + v.substring(s1);
            else
               v = key;
            input.value = v;
            return true;
         }
      }
      return false;
   },

   // API: Change event occurred.
   apiOnChange : function(dlg, id, el, e) {
      if (e.type === 'input') {
         this.contentChanged = true;
         if (this.cursorUpdated)
            this.cursorChanged(id);
      }
      else if (e.type === 'mousedown') {
         var
            t = e.target,
            p, cx, cy;
         if (t.tagName === 'INPUT') {
            if (t !== document.activeElement) {
               // See 05-util.js.
               cy = eq.Dom.parentIndex(p = t.parentNode) + 1;
               cx = eq.Dom.parentIndex(p.parentNode) + 1;
               if (this.contentChanged) {
                  this.contentChanged = false;
                  if (el.classList.contains('eq-rule')) {
                     this.cursorPending = {
                        cx : cx,
                        cy : cy
                     };
                     this.submit(el.id, el, 'mousedown');
                     return false;
                  }
               }
               this.cx = cx;
               this.cy = cy;
               this.cursorChanged(id);
               this.constructor.scrollCellIntoView(t);
            }
         }
         else if (this.selectCell(el, 0, 0))
            this.cursorChanged(id);
      }
      return true;
   },

   // API: Obtain focus element.
   apiFocusElement : function(dlg, el) {
      var csr, tab, input;
      if ((tab = eq.app.pendingTabMode) !== undefined) {
         if (el.id === eq.app.pendingTabModeId) {
            this.cursorPending = undefined;
            this.cx = this.cy = 0;
            if (tab === -1)
               input = this.prevCell(el);
            else
               input = this.nextCell(el);
         }
      }
      else if ((csr = this.cursorPending) !== undefined) {
         this.cursorPending = undefined;
         this.cx = csr.cx;
         this.cy = csr.cy;
         this.cursorUpdated = true;
      }
      if (input)
         this.cursorChanged(el.id);
      else if (!(input = this.inputCell(el)))
         input = this.nextCell(el);
      if (input)
         return input;
      this.cx = this.cy = 0;
      return el;
   },

   // API: Plugin has gained focus.
   apiFocusGained : function(dlg, ct) {
      if (ct.tagName === 'INPUT') {
         this.constructor.scrollCellIntoView(ct);
         // Select on focus except on mouse click or if focus doesn't change.
         if (   eq.app.pendingMouseDown === undefined
             && ct !== document.activeElement) {
            ct.select();
            ct.scrollLeft = ct.scrollWidth;
         }
      }
   },

   // API: Collect attribute value changes before submission.
   apiChanged : function(dlg, id, el, ty) {
      // Any attribute change requires that content is updated,
      // otherwise resuming the application could result in
      // empty Editgrid.
      this.stringAttrChanged(id, 'input', 'content', null, this.getContent(el));
      this.contentChanged = false;
   },

   // Implementation.
   // ---------------

   // Notify application: Cursor changed.
   cursorChanged : function(id) {
      this.cursorUpdated = false;
      this.intAttrChanged(id, 'change', 'cx', null, this.cx);
      this.intAttrChanged(id, 'change', 'activecolumn', null, this.cx);
      this.intAttrChanged(id, 'change', 'cy', null, this.cy);
      this.intAttrChanged(id, 'change', 'activeline', null, this.cy);
   },

   // Set number of columns.
   setColumns : function(dlg, id, el, cols) {
      var n = cols - this.cols;
      if (n) {
         var
            h = el.querySelector('.hdr>.c'),
            c = el.querySelector('.row>.c'),
            d;
         if (n > 0) {
            // Number of columns increased.
            var
               cls = this.constructor,
               r, nr = (d = c.firstElementChild) ? d.children.length : 0,
               disabled = this.disabled,
               cw, col;
            if (dlg.grid.wf !== undefined)
               cw = cls.cssWidth(dlg, 5);
            for (var i = 0; i < n; i++) {
               if (cw !== undefined) {
                  col = this.width.length;
                  // See 10-dlg.js.
                  this.setCssRule(id, cls.cssRule.width+col, [
                     ">.hdr>.c>*:nth-child("+col+")>*,",
                     ">.row>.c>*:nth-child("+col+")>*>*{width:" + cw + "}"
                  ]);
               }
               this.width.push(null);
               this.readOnly.push(false);
               this.hidden.push(false);
               h.appendChild(d = document.createElement('DIV'));
               d.className = 'cell';
               d.appendChild(document.createElement('SPAN'));
               c.appendChild(document.createElement('DIV'));
               for (r = 0; r < nr; r++)
                  cls.makeCell(c, null, null, false, disabled);
            }
         }
         else {
            // Number of columns reduced.
            this.width.length = cols+1;
            this.readOnly.length = cols;
            this.hidden.length = cols;
            for (var i = 0; i > n; i--) {
               h.removeChild(h.lastElementChild);
               c.removeChild(c.lastElementChild);
            }
         }
         this.cols = cols;
      }
   },

   // Set column title.
   setTitle : function(el, col, t) {
      if (col >= 0 && col <= this.cols) {
         var
            cell = col ? el.querySelector('.hdr>.c').children[col-1]
                       : el.querySelector('.hdr>.r'),
            cl = el.querySelector('.hdr').classList,
            empty = t.length === 0;
         // See 05-util.js.
         eq.Dom.setContent(cell.firstElementChild, t);
         if (empty) {
            // All titles empty?
            if (el.querySelector('.hdr>.r').firstElementChild.childNodes.length)
               empty = false;
            else {
               var c = el.querySelector('.hdr>.c').children;
               for (var i = 0, l = c.length; i < l; i++)
                  if (c[i].firstElementChild.childNodes.length) {
                     empty = false;
                     break;
                  }
            }
         }
         // Set/reset title bar 'empty' class.
         if (empty !== cl.contains('empty')) {
            if (empty)
               cl.add('empty');
            else
               cl.remove('empty');
         }
      }
   },

   // Set read-only columns.
   setReadOnly : function(el, col, readOnly) {
      if (col > 0 && col <= this.cols) {
         var cls = this.constructor;
         col--;
         if (readOnly !== this.readOnly[col]) {
            this.readOnly[col] = readOnly;
            var
               c = el.querySelector('.row>.c').children[col].children,
               disabled = this.disabled;
            for (var n = c.length; n > 0;)
               cls.modifyCell(c[--n], null, readOnly, disabled);
         }
      }
   },

   // Set column width.
   setWidth : function(dlg, id, el, col, w) {
      // See below, cssWidth() requires grid.wf.
      // Not present if temporary DlgElement is rendered.
      if (dlg.grid.wf !== undefined && col >= 0 && col <= this.cols) {
         var
            cls = this.constructor,
            cw = cls.cssWidth(dlg, w);
         this.width[col] = cw;
         this.hidden[col-1] = cw === 0;
         if (col === 0)
            // See 10-dlg.js.
            this.setCssRule(id, cls.cssRule.width,
               cw ? [ ">.hdr>.r>*,",
                      ">.row>.r>*>*{width:" + cw + "}" ]
                  : [ ">.hdr>.r,",
                      ">.row>.r{display:none}" ]
            );
         else
            this.setCssRule(id, cls.cssRule.width+col,
               cw ? [ ">.hdr>.c>*:nth-child("+col+")>*,",
                      ">.row>.c>*:nth-child("+col+")>*>*{width:" + cw + "}" ]
                  : [ ">.hdr>.c>*:nth-child("+col+"),",
                      ">.row>.c>*:nth-child("+col+"){display:none}" ]
            );
      }
   },

   // Get content.
   getContent : function(el) {
      var
         cls = this.constructor,
         delim = this.delim,
         r = el.querySelector('.row>.r').children,
         rl = r.length,
         c = el.querySelector('.row>.c').children,
         cl = c.length,
         cn = "",
         ri, ci, cRows;
      for (ri = 0; ri < rl; ri++) {
         cn += cls.readCell(r[ri]);
         for (ci = 0; ci < cl; ci++)
            cn += delim + cls.readCell(c[ci].children[ri]);
         cn += "\n";
      }
      return cn;
   },

   // Add row.
   addRow : function(el, row, cn) {
      if (row !== null && --row < 0)
         return;
      var
         cls = this.constructor,
         delim = this.delim,
         readOnly = this.readOnly,
         disabled = this.disabled,
         rows = cn.split('\n'),
         rows_l = rows.length,
         r = el.querySelector('.row>.r'),
         c = el.querySelector('.row>.c').children,
         cl = c.length,
         rows_i, cols, cols_l, cBefore, cRows, ci, cols_i;
      for (rows_i = 0; rows_i < rows_l; rows_i++) {
         if ((cn = rows[rows_i]).length === 0 && rows_i === rows_l-1)
            break;
         cols_l = (cols = cn.split(delim)).length;
         // Row-title.
         cBefore = undefined;
         if (row === null)
            cBefore = null;
         else if (row < (cRows = r.children).length)
            cBefore = cRows[row];
         if (cBefore !== undefined)
            cls.makeCell(r, cBefore, cols[0], true, false);
         // Columns.
         for (ci = 0; ci < cl; ci++) {
            cBefore = undefined;
            if (row === null)
               cBefore = null;
            else if (row < (cRows = c[ci].children).length)
               cBefore = cRows[row];
            if (cBefore !== undefined) {
               cols_i = ci + 1;
               cls.makeCell(c[ci], cBefore,
                  cols_i < cols_l ? cols[cols_i] : null,
                  readOnly[ci], disabled);
            }
         }
      }
   },

   // Set row.
   setRow : function(el, row, cn) {
      if (--row < 0)
         return;
      var
         cls = this.constructor,
         delim = this.delim,
         readOnly = this.readOnly,
         disabled = this.disabled,
         r = el.querySelector('.row>.r'),
         c = el.querySelector('.row>.c').children,
         cl = c.length,
         cols = cn.split('\n')[0].split(delim),
         cols_l = cols.length,
         cRows, ci, cols_i;
      // Row-title.
      if (row < (cRows = r.children).length)
         cls.modifyCell(cRows[row], cols[0], true, false);
      // Columns.
      for (ci = 0; ci < cl; ci++)
         if (row < (cRows = c[ci].children).length) {
            cols_i = ci + 1;
            cls.modifyCell(cRows[row],
               cols_i < cols_l ? cols[cols_i] : '',
               readOnly[ci], disabled);
         }
   },

   // Get row.
   getRow : function(el, row) {
      if (--row >= 0) {
         var
            cls = this.constructor,
            delim = this.delim,
            r = el.querySelector('.row>.r').children,
            c = el.querySelector('.row>.c').children,
            cl = c.length,
            cn, ci;
         if (row < r.length) {
            cn = cls.readCell(r[row]);
            for (ci = 0; ci < cl; ci++)
               cn += delim + cls.readCell(c[ci].children[row]);
            return cn;
         }
      }
      return undefined;
   },

   // Set cell.
   setCell : function(el, row, col, cn) {
      if (--row < 0 || col < 0)
         return;
      var
         cls = this.constructor,
         c;
      if (col == 0) {
         c = el.querySelector('.row>.r').children;
         if (row < c.length)
            cls.modifyCell(c[row], cn, true, false);
      }
      else {
         c = el.querySelector('.row>.c').children;
         if (--col < c.length) {
            c = c[col].children;
            if (row < c.length)
               cls.modifyCell(c[row], cn, this.readOnly[col], this.disabled);
         }
      }
   },

   // Get cell.
   getCell : function(el, row, col) {
      if (--row >= 0 && col >= 0) {
         var c;
         if (col == 0) {
            c = el.querySelector('.row>.r').children;
            if (row < c.length)
               return this.constructor.readCell(c[row]);
         }
         else {
            c = el.querySelector('.row>.c').children;
            if (--col < c.length) {
               c = c[col].children;
               if (row < c.length)
                  return this.constructor.readCell(c[row]);
            }
         }
      }
      return undefined;
   },

   // Select cell.
   selectCell : function(el, cx, cy) {
      if (cx !== this.cx || cy !== this.cy) {
         this.cx = cx;
         this.cy = cy;
         if (!this.inputCell(el))
            this.cx = this.cy = 0;
         return true;
      }
      return false;
   },

   // Obtain current input cell, null if none.
   inputCell : function(el) {
      var col, row, c, input;
      if (   (col = this.cx-1) >= 0 && (row = this.cy-1) >= 0
          && col < this.cols && !this.readOnly[col] && !this.hidden[col]) {
         c = el.querySelector('.row>.c').children;
         if (   row < (c = c[col].children).length
             && (input = c[row].firstElementChild).tagName === 'INPUT')
            return input;
      }
      return null;
   },

   // Obtain next input cell, null if none.
   nextCell : function(el) {
      if (this.cols > 0) {
         var
            readOnly = this.readOnly,
            hidden = this.hidden,
            cols = el.querySelector('.row>.c').children,
            cols_l = cols.length,
            col = this.cx-1,
            row = this.cy-1,
            rows, rows_l, input;
         if (row < 0)
            row = 0;
         while (true) {
            rows = null;
            if (++col >= cols_l) {
               rows_l = (rows = cols[col = 0].children).length;
               if (++row >= rows_l)
                  break;
            }
            if (!readOnly[col] && !hidden[col])
               break;
         }
         if (!rows)
            rows_l = (rows = cols[col].children).length;
         if (   row >= 0 && row < rows_l
             && (input = rows[row].firstElementChild).tagName === 'INPUT') {
            this.cx = col+1;
            this.cy = row+1;
            return input;
         }
      }
      return null;
   },

   // Obtain previous input cell, null if none.
   prevCell : function(el) {
      if (this.cols > 0) {
         var
            readOnly = this.readOnly,
            hidden = this.hidden,
            cols = el.querySelector('.row>.c').children,
            cols_l = cols.length,
            col = this.cx-1,
            row = this.cy-1,
            rows, rows_l, input;
         if (row < 0)
            row = cols[0].children.length;
         while (true) {
            rows = null;
            if (--col < 0) {
               rows_l = (rows = cols[col = cols_l-1].children).length;
               if (--row < 0)
                  break;
            }
            if (!readOnly[col] && !hidden[col])
               break;
         }
         if (!rows)
            rows_l = (rows = cols[col].children).length;
         if (   row >= 0 && row < rows_l
             && (input = rows[row].firstElementChild).tagName === 'INPUT') {
            this.cx = col+1;
            this.cy = row+1;
            return input;
         }
      }
      return null;
   },

   // Static methods.
   // ---------------
   static : {

      // Static: Clear rows.
      clearRows : function(el) {
         // Row-titles.
         el.querySelector('.row>.r').textContent = ''; // Delete children.
         // Columns.
         var c = el.querySelector('.row>.c').children;
         for (var i = 0, l = c.length; i < l; i++)
            c[i].textContent = ''; // Delete children.
      },

      // Static: Delete rows.
      delRows : function(el, row, del) {
         if (--row < 0)
            return;
         // Row-titles.
         var
            p = el.querySelector('.row>.r'),
            r = p.children,
            n, c;
         for (n = del; n-- > 0;)
            if (row < r.length)
               p.removeChild(r[row]);
         // Columns.
         c = el.querySelector('.row>.c').children;
         for (var i = 0, l = c.length; i < l; i++) {
            r = (p = c[i]).children;
            for (n = del; n-- > 0;)
               if (row < r.length)
                  p.removeChild(r[row]);
         }
      },

      // Static: Make cell.
      makeCell : function(parent, before, val, readOnly, disabled) {
         var cell, input, span;
         parent.insertBefore(cell = document.createElement('DIV'), before);
         cell.className = 'cell';
         if (!readOnly) {
            cell.appendChild(input = document.createElement('INPUT'));
            input.type = 'text';
            input.setAttribute('autocomplete', 'off');
            input.setAttribute('autocorrect', 'off');
            input.setAttribute('spellcheck', false);
            if (val)
               input.value = val;
            if (disabled)
               input.disabled = disabled;
         }
         else {
            cell.appendChild(span = document.createElement('SPAN'));
            if (val)
               // See 05-util.js.
               eq.Dom.setContent(span, val);
         }
      },

      // Static: Modify cell.
      modifyCell : function(cell, val, readOnly, disabled) {
         var input = cell.firstElementChild, span;
         if (input.tagName !== 'INPUT') {
            span = input;
            input = null;
         }
         if (readOnly) {
            if (input) {
               if (val === null)
                  val = input.value;
               cell.removeChild(input);
            }
            if (!span)
               cell.appendChild(span = document.createElement('SPAN'));
            if (val !== null)
               // See 05-util.js.
               eq.Dom.setContent(span, val);
         }
         else {
            if (!input) {
               if (val === null)
                  val = cell.textContent;
               cell.textContent = ''; // Delete children.
               cell.appendChild(input = document.createElement('INPUT'));
               input.type = 'text';
               input.setAttribute('autocomplete', 'off');
               input.setAttribute('autocorrect', 'off');
               input.setAttribute('spellcheck', false);
            }
            if (val !== null)
               input.value = val;
            if (disabled)
               input.disabled = disabled;
         }
      },

      // Static: Read cell.
      readCell : function(cell) {
         var input = cell.firstElementChild;
         if (input.tagName === 'INPUT')
            return input.value;
         return cell.textContent;
      },

      // Static: Scroll cell into view.
      scrollCellIntoView : function(ct) {
         if (ct.tagName === 'INPUT') {
            ct = ct.parentNode;
            var
               rc = ct.getBoundingClientRect(),
               pane = ct.parentNode.parentNode,
               rcPane = pane.getBoundingClientRect(),
               v0, v1;
            if ((v0 = rc.top) < (v1 = rcPane.top))
               pane.scrollTop = Math.floor(pane.scrollTop - v1 + v0);
            else if ((v0 = rc.bottom) > (v1 = rcPane.bottom))
               pane.scrollTop = Math.floor(pane.scrollTop + v0 - v1);
            if ((v0 = rc.left) < (v1 = rcPane.left))
               pane.scrollLeft = Math.floor(pane.scrollLeft - v1 + v0);
            else if ((v0 = rc.right) > (v1 = rcPane.right))
               pane.scrollLeft = Math.floor(pane.scrollLeft + v0 - v1);
         }
      },

      // Static: Return width/height in CSS units.
      cssWidth : function(dlg, w) {
         // Layout grid width ch factor, see 10-dlg.js.
         if (w)
            return (Math.round(w * dlg.grid.wf * 1000) / 1000) + 'ch';
         return 0;
      },
      cssHeight : function(dlg, h) {
         // Layout grid height em factor, see 10-dlg.js.
         if (h)
            return (Math.round(h * dlg.grid.hf * 1000) / 1000) + 'em';
         return 0;
      },

      // Static event handler: Columns scrolled.
      onScrollColumns : function(e) {
         // See 05-util.js.
         eq.Dom.onNextFrame(this, {
            cb : eq.c.plgEditgrid.onScrollColumns_exec
         });
      },
      onScrollColumns_exec : function(el) {
         // Scroll column titles accordingly.
         var t = el.parentNode.previousElementSibling.lastElementChild
               .firstElementChild;
         if (t)
            t.style.marginLeft = '-' + el.scrollLeft + 'px';
         // Scroll row titles accordingly.
         t = el.previousElementSibling.firstElementChild;
         if (t)
            t.style.marginTop = '-' + el.scrollTop + 'px';
      },

      cssRule : {
         font   : 101,
         height : 102,
         width  : 110
      }
   }
});
